
"""
Functions that print to the console with optional sentiment using marks.
"""

import typing
import os

from . import _constants
from . import _helpers
from . import _system
from . import _colors
from . import _theme


__all__ = ('text', 'info', 'done', 'fail')


_type_text_sep = str
_type_text_end = str
_type_text_re  = bool


def _text(*values, 
          sep: _type_text_sep = ' ', 
          end: _type_text_end = _constants.linesep, 
          re : _type_text_re  = False):

    """
    Print a plain value. Works similar to :func:`print`.

    :param values:
        Used to create the final value by mapping all with :func:`str` and concatenating with :paramref:`.text.sep`.
    :param end:
        Appended to the end of the final value.
    :param re:
        Whether to return the cursor to the last saved position before printing.
    """

    def sketch():
        value = sep.join(map(str, values)) + end
        lines = _helpers.split_lines(value)
        point = None
        return (lines, point)

    _system.screen.print(sketch, re)


text = _text


def _annotate(mark_color, mark, values, **kwargs):

    if not mark_color is None:
        mark = _helpers.paint_text(mark_color, mark)

    _text(mark, *values, **kwargs)


_type_info_mark       = str
_type_info_mark_color = str


@_theme.add('printers.info')
def _info(*values,
          mark      : _type_info_mark       = '!', 
          mark_color: _type_info_mark_color = _colors.basic('cyan'),
          **kwargs):

    """
    Print a value denoting information.

    :param values:
        Same as :paramref:`.text.value`.
    :param mark:
        Prepended to the final value.
    :param mark_color:
        Used to color :paramref:`.info.mark` with.

    Additional arguments are passed to :func:`.text`.

    |theme| :code:`printers.info`.

    .. image:: /_static/images/printers.info-1.png
    """

    _annotate(mark_color, mark, values, **kwargs)


info = _info


_type_done_mark       = str
_type_done_mark_color = str


@_theme.add('printers.done')
def _done(*values,
          mark      : _type_done_mark       = '!' if os.name == 'nt' else '✔', 
          mark_color: _type_done_mark_color = _colors.basic('green'),
          **kwargs):

    """
    Print a value denoting success.

    :param values:
        Same as :paramref:`.text.value`.
    :param mark:
        Prepended to the final value.
    :param mark_color:
        Used to color :paramref:`.info.mark` with.

    Additional arguments are passed to :func:`.text`.

    |theme| :code:`printers.done`.

    .. image:: /_static/images/printers.done-1.png
    """

    _annotate(mark_color, mark, values, **kwargs)


done = _done


_type_fail_mark       = str
_type_fail_mark_color = str


@_theme.add('printers.fail')
def _fail(*values,
          mark      : _type_fail_mark       = '!' if os.name == 'nt' else '✘', 
          mark_color: _type_fail_mark_color = _colors.basic('red'),
          **kwargs):

    """
    Print a value denoting failure.

    :param values:
        Same as :paramref:`.text.value`.
    :param mark:
        Prepended to the final value.
    :param mark_color:
        Used to color :paramref:`.info.mark` with.

    Additional arguments are passed to :func:`.text`.

    |theme| :code:`printers.fail`.

    .. image:: /_static/images/printers.fail-1.png
    """

    _annotate(mark_color, mark, values, **kwargs)


fail = _fail




"""
Classes for changing data in-place in ways that emulate commonly-used interfaces.
"""

import typing
import abc
import types
import copy

from . import _helpers


__all__ = ('Error', 'Mutate', 'Cursor', 'Text', 'Mesh')


class Error(_helpers.InfoErrorMixin, Exception):

    __slots__ = ()


class Mutate(abc.ABC):

    """
    Base class for mutates. They emulate the behavior of common data manipulators.
    """

    __slots__ = ()

    class _State(types.SimpleNamespace):

        __slots__ = ()

    @abc.abstractmethod
    def _get_state(self):

        return NotImplemented
    
    def get_state(self):

        """
        Get a copy of the current state.
        """

        return self._get_state()
    
    @abc.abstractmethod
    def _set_state(self, state):

        return NotImplemented
    
    def set_state(self, state):

        """
        Restore the current state to the one provided.
        """

        return self._set_state(state)
    

_type_cursor_point = typing.List[int]
    

class Cursor(Mutate):

    """
    A mutate for a point representing a position in space.

    :param point:
        The manipulating point.
    """

    __slots__ = ('_point',)

    def __init__(self, 
                 point: _type_cursor_point):

        self._point = point

    @property
    def point(self) -> _type_cursor_point:

        """
        The internal point.
        """

        return self._point
    
    def _get_state(self):
        
        point = copy.copy(self._point)

        return self._State(point = point)
    
    def _set_state(self, state):

        self._point[:] = state.point
    
    def _move(self, instructions):

        for axis, coordinate in instructions:
            self._point[axis] = coordinate

    def move(self, 
             instructions: typing.List[typing.Tuple[int, int]]) -> None:

        """
        Move according to the instructions given. 
        
        Each instruction should be of ``(index, new_coordinate)`` form.
        """

        self._move(instructions)


_type_Text_init_rune  = str
_type_Text_init_line  = typing.List[_type_Text_init_rune]
_type_Text_init_lines = typing.List[_type_Text_init_line]
_type_Text_init_point = _type_cursor_point


class Text(Mutate):

    """
    A mutate for a block of text and a cursor among it.

    :param lines:
        The manipulating list of lines.
    :param point:
        The point for the cursor.
    """

    __slots__ = ('_lines', '_cursor')

    def __init__(self, 
                 lines: _type_Text_init_lines, 
                 point: _type_Text_init_point):

        self._lines = lines
        self._cursor = Cursor(point)

    @property
    def lines(self) -> _type_Text_init_lines:

        """
        The current lines.
        """

        return self._lines

    @property
    def _point(self):

        return self._cursor.point

    @property
    def point(self) -> _type_Text_init_point:

        """
        The current cursor's point.
        """

        return self._point
    
    def _get_state(self):

        lines = copy.deepcopy(self._lines)

        cursor = self._cursor.get_state()

        return self._State(
            lines = lines,
            cursor = cursor
        )
    
    def _set_state(self, state):

        self._lines[:] = state.lines

        self._cursor.set_state(state.cursor)

    def _insert(self, runes):

        cur_y = self._point[0]
        
        cur_x = self._point[1]

        cur_lines = self._lines
        
        cur_line = cur_lines[cur_y]

        cur_line[cur_x:cur_x] = runes

    def insert(self, 
               runes: typing.List[str]):

        """
        Insert text at the current cursor position and move it accordingly.

        :param runes:
            The runes to insert.
        """

        self._insert(runes)

    def _move_y(self, size):

        cur_y = self._point[0]
        cur_x = self._point[1]
        cur_lines = self._lines
        
        new_y = cur_y + size

        max_y = len(cur_lines) - 1
        
        if not 0 <= new_y <= max_y:
            raise Error('insufficient_y_space', size = size)

        new_line = cur_lines[new_y]

        instructions = [(0, new_y)]

        may_x = len(new_line)

        if cur_x > may_x:
            instructions.append((1, may_x))

        self._cursor.move(instructions)

    def move_y(self, 
               size: int):
        
        """
        Move the cursor verticaly, wrapping to the max width of the new line as necessary.

        :param size:
            The amount of lines to move by.
        """

        self._move_y(size)

    def _move_x(self, size):

        cur_y = self._point[0]
        cur_x = self._point[1]
        cur_lines = self._lines

        cur_i = _helpers.text_point_to_index(cur_lines, cur_y, cur_x)

        min_i = 0
        max_i = sum(map(len, cur_lines)) + len(cur_lines) - 1

        new_i = cur_i + size

        if not min_i <= new_i <= max_i:
            raise Error('insufficient_x_space', size = size)

        new_y, new_x = _helpers.text_index_to_point(cur_lines, new_i)

        instructions = ((0, new_y), (1, new_x))

        self._cursor.move(instructions)

    def move_x(self, 
               size: int):

        """
        Move the cursor horizontally, wrapping to the next or last line as necessary.

        :param size:
            The amount of runes to move by.
        """

        self._move_x(size)

    def _delete(self, size):

        cur_y = self._point[0]
        cur_x = self._point[1]
        
        cur_lines = self._lines
        cur_line = cur_lines[cur_y]

        cur_i = _helpers.text_point_to_index(cur_lines, cur_y, cur_x)

        max_i = sum(map(len, cur_lines)) + len(cur_lines) - 1

        new_i = cur_i + size

        if not new_i <= max_i:
            raise Error('insufficient_x_space', size = size)

        new_y, new_x = _helpers.text_index_to_point(cur_lines, new_i)

        if not cur_y == new_y:
            psh_y = cur_y + 1
            cur_lines[psh_y:new_y] = ()
            mrg_line = cur_lines.pop(psh_y)
            new_x += len(cur_line)
            cur_line.extend(mrg_line)

        cur_line[cur_x:new_x] = ()

    def delete(self, 
               size: int):

        """
        Delete ahead of the cursor, including newlines.

        :param size:
            The amount of runes to delete.
        """

        self._delete(size)

    def _newline(self):

        cur_y = self._point[0]
        cur_x = self._point[1]
        
        cur_lines = self._lines
        cur_line = cur_lines[cur_y]

        new_line = cur_line[cur_x:]

        cur_line[cur_x:] = ()

        psh_y = cur_y + 1

        cur_lines.insert(psh_y, new_line)

        instructions = ((0, psh_y), (1, 0))

        self._cursor.move(instructions)

    def newline(self):

        """
        Insert a new line under the cursor, cutting the current one as necessary.
        """

        self._newline()


_type_Mesh_init_clean  = typing.Union[int, None]
_type_Mesh_init_spot   = typing.Tuple[int, int]
_type_Mesh_init_tile   = typing.Any
_type_Mesh_init_tiles  = typing.Dict[_type_Mesh_init_spot, _type_Mesh_init_tile]
_type_Mesh_init_point  = _type_cursor_point
_type_Mesh_init_score  = typing.Union[typing.Callable[[_type_Text_init_lines, _type_Mesh_init_tile], typing.Union[int, None]], None]
_type_Mesh_init_scout  = typing.Union[typing.Callable[[_type_Mesh_init_spot], bool], None]
_type_Mesh_init_rigid  = bool
_type_Mesh_init_create = typing.Union[typing.Callable[[_type_Mesh_init_spot], typing.Union[_type_Mesh_init_tile, None]], None]


class Mesh(Mutate):

    """
    A mutate for a collection of tiles.

    :param score:
        Used during searching to determine the order at which each spot's tile should be shown, if at all.
    :param scout:
        Determines whether a spot is valid to move to.
    :param rigid:
        Whether to **not** allow wraping around when moving toward the edge.
    :param create:
        Used for creating tiles for new spots. Can return :code:`None` to prevent creation.
    :param tiles:
        The manipulating tiles.
    :param point:
        The point for the cursor.
    """

    __slots__ = ('_create', '_scout', '_rigid', '_search_score', '_search_mutate', 
                 '_vision', '_search_point_cache', '_clean', '_tiles', '_cursor')

    def __init__(self, 
                 score : _type_Mesh_init_score, 
                 scout : _type_Mesh_init_scout, 
                 rigid : _type_Mesh_init_rigid,
                 create: _type_Mesh_init_create, 
                 clean : _type_Mesh_init_clean,
                 tiles : _type_Mesh_init_tiles, 
                 point : _type_Mesh_init_point,
                 *args, **kwargs):

        self._create = create

        self._scout = scout

        self._rigid = rigid

        search_lines = [[]]

        search_point_y = len(search_lines) - 1
        search_point_x = len(search_lines[search_point_y])
        search_point = [search_point_y, search_point_x]

        self._search_mutate = Text(search_lines, search_point)
        self._search_score = score
        self._vision = {spot: spot for spot in tiles}
        self._search_point_cache = None

        self._clean = clean

        self._tiles = tiles
        
        self._cursor = Cursor(point)

        super().__init__(*args, **kwargs)

        self._insert(self._vis_spot)
    
    @property
    def _search_point(self):

        return self._search_mutate.point
    
    @property
    def search_point(self) -> _type_Mesh_init_spot:

        """
        The search mutate's cursor's point.
        
        Should always be at the end of the :attr:`.search_line`.
        """

        return self._search_point
    
    @property
    def _search_lines(self):

        return self._search_mutate.lines
    
    @property
    def search_lines(self) -> _type_Text_init_lines:

        """
        The search mutate's lines.

        Only one line will be in use at any given time.
        """

        return self._search_lines

    @property
    def _search_line(self):

        return self._search_lines[self._search_point[0]]

    @property
    def search_line(self) -> _type_Text_init_line:

        """
        The search mutate's current line.
        
        Tile creation is only possible when this is empty.
        """

        return self._search_line
    
    @property
    def vision(self) -> typing.Dict[_type_Mesh_init_spot, _type_Mesh_init_spot]:

        """
        A mapping of currently visible spots to the real ones they correspond to.
        """

        return self._vision

    @property
    def tiles(self) -> _type_Mesh_init_tiles:

        return self._tiles

    @property
    def _point(self):

        return self._cursor.point

    @property
    def point(self) -> _type_Mesh_init_point:

        """
        The current cursor's point.
        """

        return self._point

    @property
    def _vis_spot(self):

        return tuple(self._point)
    
    @property
    def vis_spot(self) -> _type_Mesh_init_spot:

        """
        The current vision spot. Simply a tuple of :attr:`.point`.
        """

        return self._vis_spot
    
    @property
    def _cur_spot(self):

        spot = self._vision[self._vis_spot]

        return spot
    
    @property
    def cur_spot(self) -> _type_Mesh_init_spot:

        """
        The current real spot, accounting for vision.
        """

        return self._cur_spot
    
    @property   
    def _cur_tile(self):

        return self._tiles[self._cur_spot]
    
    @property
    def cur_tile(self):

        """
        The current tile.
        """

        return self._cur_tile
    
    def _get_state(self):

        tiles = copy.copy(self._tiles)
        
        cursor = self._cursor.get_state()

        search = self._search_mutate.get_state()
        vision = copy.deepcopy(self._vision)

        return self._State(
            tiles = tiles, 
            cursor = cursor, 
            search = search, 
            vision = vision
        )
    
    def _set_state(self, state):

        self._tiles.clear()
        self._tiles.update(state.tiles)
        
        self._cursor.set_state(state.cursor)

        self._search_mutate.set_state(state.search)
        self._vision.clear()
        self._vision.update(state.vision)
    
    def _insert(self, spot):

        if self._search_line:
            raise Error('searching')

        try:
            tile = self._tiles[spot]
        except KeyError:
            pass
        else:
            return tile

        tile = self._create(spot)

        if self._clean:
            self._tiles.clear()
            self._vision.clear()

        if not tile is None:
            self._tiles[spot] = tile
            self._vision[spot] = spot

        return tile
    
    def insert(self, 
               spot: _type_Mesh_init_spot) -> typing.Union[_type_Mesh_init_tile, None]:
        
        """
        Attempt to insert a tile.

        :param spot:
            The spot to insert the tile at.

        If one already exists or gets created, it is returned.
        """

        return self._insert(spot)
    
    def _delete(self, spot):

        if self._search_line:
            raise Error('searching')
        
        try:
            tile = self._tiles.pop(spot)
        except KeyError:
            return

        try:
            del self._vision[spot]
        except KeyError:
            pass

        return tile

    def delete(self,
               spot: _type_Mesh_init_spot) -> typing.Union[_type_Mesh_init_tile, None]:

        """
        Attempt to delete a tile
        
        :param spot:
            The spot to delete the tile from.
        
        If one exists, it is returned.
        """    

        return self._delete(spot)
    
    def _move(self, direction, size = 1):

        old_spot = self._vis_spot

        fin_spot = None
        if not self._search_line:
            if not self._create is None:
                new_spot = _helpers.get_point_directional_next(old_spot, direction, size)
                if not new_spot in self._tiles:
                    tile = self._insert(new_spot)
                    if not tile is None:
                        fin_spot = new_spot

        if fin_spot is None:
            may_spots = self._vision.keys()
            if not self._scout is None:
                may_spots = filter(self._scout, may_spots) 
            may_spots = tuple(may_spots)
            fin_spot = _helpers.get_point_directional(min, old_spot, direction, may_spots)
            if fin_spot == old_spot:
                if self._rigid:
                    raise Error('insufficient_space', direction = direction)
                direction = _helpers.reverse_direction(direction)
                fin_spot = _helpers.get_point_directional(max, old_spot, direction, may_spots, ignore_point = True)
        
        instructions = enumerate(fin_spot)

        self._cursor.move(instructions)

    def move(self, 
             direction: int):

        """
        Move the cursor to the nearest tile.
         
        :param direction:
            The direction to move toward in circular ``[-90, +180]`` degrees.

        If none exists toward the specified direction, attempt again by wrapping around the other side.
        """

        self._move(direction)

    def _search_execute_setup(self, argument):

        if self._search_point_cache is None:
            self._search_point_cache = copy.copy(self._point)

        assets = []
        for spot, tile in self._tiles.items():
            score = self._search_score(argument, tile)
            if score is None:
                continue
            assets.append((score, spot))

        if not assets:
            raise Error('invalid_search_argument', argument = argument)
        
        assets = sorted(assets, reverse = True)
        
        scores, old_spots = zip(*assets)

        cur_spots = self._vision.values()

        if set(old_spots) == set(cur_spots):
            raise Error('inconsequential_search_argument', argument = argument)
            
        combos = _helpers.squeeze_spots(0, old_spots)

        self._vision = dict(map(reversed, combos))

        return next(iter(self._vision))

    def _search_execute_reset(self):

        self._vision = {spot: spot for spot in self._tiles}

        point = self._search_point_cache
        self._search_point_cache = None
        
        return point

    def _search_execute(self, function, *args, **kwargs):

        if self._search_score is None:
            return

        argument = self._search_line

        if argument is None:
            return

        state = self._get_state()

        function(*args, **kwargs)

        try:
            if argument:
                point = self._search_execute_setup(argument)
            else:
                point = self._search_execute_reset()
        except Error:
            self._set_state(state)
            raise

        instructions = enumerate(point)

        self._cursor.move(instructions)

    def _search_insert_act(self, runes):

        self._search_mutate.insert(runes)

        size = len(runes)

        self._search_mutate.move_x(size)

    def _search_insert(self, runes):

        self._search_execute(self._search_insert_act, runes)

    def search_insert(self, 
                      runes: typing.List[str]):

        """
        Insert text to the search line, and filter or displace tiles accordingly.

        :param runes:
            The runes to insert.
        """

        self._search_insert(runes)

    def _search_delete_act(self, size):

        self._search_mutate.move_x(- size)
        self._search_mutate.delete(size)

    def _search_delete(self, size):

        self._search_execute(self._search_delete_act, size)

    def search_delete(self, 
                      size: int):

        """
        Delete from the search line, and filter or displace tiles accordingly.

        :param size:
            The amount of runes to delete.
        """

        self._search_delete(size)



"""
Core assets used throughout the module that read and write to the system io buffers.
"""

import sys

from . import _core


__all__ = ('io', 'cursor', 'render', 'handle', 'screen', 'console')


#:
_io = io = _core.IO(sys.stdin, sys.stdout)

#:
_intel = intel = _core.Intel(_io)

#:
_handle = handle = _core.Handle(_intel)
 
#:
_cursor = cursor = _core.Cursor(_intel)

#:
_render = render = _core.Render(_cursor)

#:
_screen = screen = _core.Screen(_render)

#:
_console = console = _core.Console(_handle, _screen)





"""
Themeing is essentially changing the default values for elligible callables.

Look for the "|theme|" sections to see which functions are subject to themeing. 
"""

import functools
import contextlib


__all__ = ('use', 'get')


_store = []

_store_default = {}


@contextlib.contextmanager
def _use(theme: dict):

    """
    Use the theme where applicable. This overrides any subsequent themes.

    A theme is a collection of default arguments for specified functions.

    .. code-block:: python

        with theme.use({'printers.info': {'mark_color': colors.basic('magenta')}}):
            with theme.use({'printers.done': {'mark_color': colors.basic('yellow')}}):
                survey.printers.info('example')
                survey.printers.done('example')

    On the above example, only the marker of ``info`` will be painted magenta.
    """

    _store.append(theme)

    yield

    _store.pop()


use = _use


def _get():

    """
    Get the current theme, with all subsequent values attached to it.
    """

    try:
        return _store[0]
    except IndexError:
        return _store_default


get = _get


def add(name: str):

    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            theme = _get()
            try:
                defaults = theme[name]
            except KeyError:
                pass
            else:
                kwargs = {**defaults, **kwargs}
            return function(*args, **kwargs)
        return wrapper

    return decorator


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
import os
import importlib.metadata

sys.path.insert(0, os.path.abspath('..'))


project = 'survey'
author = 'Exahilosys'
copyright = f'2023, {author}'
release = importlib.metadata.version(project)

rst_prolog = """
.. |theme| replace:: 🎨 Theme with
"""

extensions = [
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_autodoc_typehints',
    'sphinx_paramlinks',
]

autodoc_member_order = 'bysource'

autodoc_default_options = {
    'show-inheritance': True
}

intersphinx_mapping = {
    'py': ('https://docs.python.org/3', None),
}

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

paramlinks_hyperlink_param = 'name'

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']




"""
Classes for printing data that can change over time.
"""

import typing
import time
import math
import threading
import functools
import collections
import collections.abc
import copy

from . import _core
from . import _helpers
from . import _system
from . import _visuals
from . import _funnels
from . import _colors
from . import _theme


__all__ = ('SpinProgress', 'MultiLineProgress', 'MultiLineProgressControl',
           'LineProgress')


_type_init_Interface_frequency = float


class Interface:
    
    """
    The graphics manager, responsible for printing asynchronously.

    :param frequency:
        The "refresh rate" for all graphics.
        
    .. code-block:: python
    
        with Interface() as interface:
            interface.tiles.append(...)
    """

    __slots__ = ('_frequency', '_io', '_cursor', '_render', '_screen', 
                 '_visual_tiles', '_visual_point', '_visual', '_active', 
                 '_thread')

    @_theme.add('graphics.Interface')
    def __init__(self, 
                 frequency: _type_init_Interface_frequency):
        
        self._frequency = frequency

        io = _system.io
        cursor = _system.cursor
        render = _core.Render(cursor)
        screen = _core.Screen(render)

        self._io = io
        self._cursor = cursor
        self._render = render
        self._screen = screen

        visual_funnel_enter_group = []
        visual_funnel_enter_entry = _funnels.line_delimit('\n')
        visual_funnel_enter_group.append(visual_funnel_enter_entry)
        visual_funnel_enter = _helpers.chain_functions(*visual_funnel_enter_group)

        visual_tiles = []
        visual_point = [0, 0]
        visual = _visuals.Line.link(visual_tiles, visual_point, funnel_enter = visual_funnel_enter)

        self._visual_tiles = visual_tiles
        self._visual_point = visual_point
        self._visual = visual

        self._active = True
        self._thread = None

    @property
    def tiles(self) -> typing.List[_visuals.Visual]:
        
        """
        The managed graphics.
        
        Can be modified in any way at any time.
        """

        return self._visual_tiles
    
    def _sketch(self, fin):

        lines, point = self._visual.get()

        if fin:
            lines.append([])

        return (lines, None)

    def _print(self, fin = False):

        sketch = functools.partial(self._sketch, fin)

        self._screen.print(sketch, True)

    def _cycle(self):

        while self._active:
            self._print()
            time.sleep(self._frequency)

    def _start(self):

        self._cursor.hidden.enter()

        thread = threading.Thread(target = self._cycle, daemon = True)

        thread.start()

        self._thread = thread

    def start(self):

        self._start()

    def _close(self):

        self._active = False

        if self._thread is None:
            return
        
        self._thread.join()

        self._print(fin = True)

        self._cursor.hidden.leave()

    def close(self):

        self._close()
        
        
_type_init_Visual_get      = _visuals._type_Text_init_get
_type_init_Visual_epoligue = typing.Union[str, typing.Callable[['Visual'], str]]

        
class Visual(_visuals.Text):
    
    """
    A base class that communicates with an :class:`.Interface` upon creation and deletion.
    
    :param get:
        Same as :paramref:`.visuals.Text.get`.
    :param epilogue:
        Used for the final print before closing.
    """

    _interface = None
    _interface_frequency = NotImplemented

    __slots__ = ('_epilogue', '_phantasm')

    def __init__(self, 
                 get     : _type_init_Visual_get, 
                 *args, 
                 epilogue: _type_init_Visual_epoligue = None,
                 **kwargs):

        self._epilogue = epilogue
        self._phantasm = None

        sub_get = get

        def get(*args, **kwargs):
            lines = self._phantasm
            if lines is None:
                value = sub_get(*args, **kwargs)
                lines = _helpers.split_lines(value)
            else:
                lines = copy.deepcopy(lines)
            point = [0, 0]
            return (lines, point)
        
        super().__init__(get, *args, **kwargs)

    def _start(self):

        init = self._interface is None

        if init:
            interface = self.__class__._interface = Interface(self._interface_frequency)
        else:
            interface = self._interface

        interface.tiles.append(self.get)

        if init:
            interface.start()

    def _close(self):
        
        epilogue = _helpers.get_or_call(self._epilogue, self)

        self._phantasm = self._get()[0] if epilogue is None else _helpers.split_lines(epilogue)

        interface = self._interface

        if any(get.__self__._phantasm is None for get in self._interface.tiles):
            return
        
        self.__class__._interface = None

        interface.close()

    def __enter__(self, *args):

        self._start()

        return self

    def __exit__(self, *args):

        self._close()
        

_type_init_Graphic_get        = typing.Union[str, typing.Callable[['Visual'], str]]
_type_init_Graphic_prefix     = typing.Union[str, typing.Callable[['Visual'], str]]
_type_init_Graphic_suffix     = typing.Union[str, typing.Callable[['Visual'], str]]
_type_init_Graphic_mark       = str
_type_init_Graphic_mark_color = str


class Graphic(Visual):
    
    """
    A base class for graphics.
    
    :param get:
        Same as :paramref:`.Visual.get`, except it should return `str`.
    :param prefix:
        Added before the result of :paramref:`.get`.
    :param suffix:
        Added after the result of :paramref:`.get`.
    :param mark:
        Added before :paramref:`.prefix`.
    :param mark_color:
        The color to paint :paramref:`.mark` with.
    """

    __slots__ = ('_info',)

    def __init__(self,
                 get       : _type_init_Graphic_get,
                 *args,
                 prefix    : _type_init_Graphic_prefix     = None,
                 suffix    : _type_init_Graphic_suffix     = None,
                 mark      : _type_init_Graphic_mark       = '! ', 
                 mark_color: _type_init_Graphic_mark_color = _colors.basic('cyan'),
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]

        epilogue = _helpers.get_function_arg_safe(super_cls, 'epilogue', kwargs, pop = True)

        mark = _helpers.paint_text(mark_color, mark)

        def merge_parts(parts):
            if not mark is None:
                parts.insert(0, mark)
            return ''.join(parts)

        def epilogue(*args, __sub = epilogue, **kwargs):
            value = _helpers.get_or_call(__sub, *args, **kwargs)
            if value is None:
                return value
            parts = [value]
            return merge_parts(parts)

        if not callable(prefix):
            sub_prefix = prefix
            prefix = lambda *args: sub_prefix

        if not callable(suffix):
            sub_suffix = suffix
            suffix = lambda *args: sub_suffix

        sub_get = get
        def get(*args):
            value = sub_get(*args)
            parts = []
            value_prefix = prefix(self)
            value_suffix = suffix(self)
            if not value_prefix is None:
                parts.append(value_prefix)
            parts.append(value)
            if not value_suffix is None:
                parts.append(value_suffix)
            return merge_parts(parts)
        
        super().__init__(get, *args, epilogue = epilogue, **kwargs)


class SpinProgress(Graphic):
    
    """
    A spinner indicating work of undeterminable size.
    
    :param runes:
        The characters to cycle through on every print.

    .. code-block:: python

        state = None
        with survey.graphics.SpinProgress(prefix = 'Loading ', suffix = lambda self: state, epilogue = 'Done!') as progress:
            for state in (state, ' calculating...', ' molding...', ' redering...'):
                time.sleep(2)
    
    |theme| :code:`graphics.SpinProgress`.

    .. image:: /_static/images/graphics.spinprogress-1.gif
    """

    _interface_frequency = 0.2

    __slots__ = ('_stage',)

    @_theme.add('graphics.SpinProgress')
    def __init__(self,
                 *args,
                 runes: typing.List[str] = ('|', '/', '—', '\\'),
                 **kwargs):

        self._stage = 0

        def get(*args):
            index = self._stage % len(runes)
            rune = runes[index]
            self._stage += 1
            return rune
        
        super().__init__(get, *args, **kwargs)


_type_MultiLineProgressControl_init_total      = typing.Union[int, float]
_type_MultiLineProgressControl_init_value      = typing.Union[int, float]
_type_MultiLineProgressControl_init_runes      = typing.List[str]
_type_MultiLineProgressControl_init_color      = typing.Union[str, None]
_type_MultiLineProgressControl_init_suffix     = typing.Union[str, typing.Callable[['MultiLineProgressControl'], str]]
_type_MultiLineProgressControl_init_epilogue   = typing.Union[str, typing.Callable[['MultiLineProgressControl'], str]]
_type_MultiLineProgressControl_init_denominate = typing.Union[typing.Callable[[_type_MultiLineProgressControl_init_value], typing.Tuple[float, str]], None]

_type_MultiLineProgressControl_move_size = _type_MultiLineProgressControl_init_value

_type_MultiLineProgressControl_get_line_width = int


class MultiLineProgressControl:
    
    """
    A line indicating the progress of work of a certain size.
    
    :param total:
        The maximum expected amount of the internal value.
    :param value:
        The initial value.
    :param runes:
        The runes to cycle through for subdivisions of a character.
    :param color:
        The color of the line.
    :param suffix:
        Added after the line. The following placeholders may be used:

        - ``value`` is the current value. 
        - ``total`` is :paramref:`.total`.
        - ``total_unit`` is from :paramref:`.denominate` on ``total``.
        - ``speed`` is the known moving rate of the value. 
        - ``speed_unit`` is from :paramref:`.denominate` on ``speed``.
        - ``remain`` is the known time left until completion. 
        - ``elapse`` is the time spent since the first move.
    :param epilogue:
        Used after the value has reached the total instead of the suffix.
    :param denominate:
        Used to determine the desired denomination of the value (and speed) and its respective unit.

        For example, if speed is :code:`'1000 KB/s'`, then :code:`denominate(1000)` may return :code:`(1000, 'MB')` so that speed is shown as :code:`1000/1000 -> '1 MB/s'` 
        
    |theme| :code:`.graphics.MultiLineProgressControl`
    """
    
    __slots__ = ('_total', '_value', '_runes', '_color', '_suffix', 
                 '_epilogue', '_denominate', '_init_time', '_last_time')

    @_theme.add('graphics.MultiLineProgressControl')
    def __init__(self, 
                 total     : _type_MultiLineProgressControl_init_total,
                 value     : _type_MultiLineProgressControl_init_value      = 0,
                 runes     : _type_MultiLineProgressControl_init_runes      = ('━',),
                 color     : _type_MultiLineProgressControl_init_color      = None,
                 suffix    : _type_MultiLineProgressControl_init_suffix     = '{value}/{total}{total_unit} {speed}{speed_unit}/s {remain}',
                 epilogue  : _type_MultiLineProgressControl_init_epilogue   = None,
                 denominate: _type_MultiLineProgressControl_init_denominate = None):
        
        if denominate is None:
            denominate = lambda value: (1, '')

        self._total = total
        self._value = value

        self._runes = runes
        self._color = color
        self._suffix = suffix
        self._epilogue = epilogue
        self._denominate = denominate

        self._init_time = None
        self._last_time = None

    @property
    def total(self) -> _type_MultiLineProgressControl_init_total:

        """
        The total value.
        """

        return self._total
        
    @property
    def value(self) -> _type_MultiLineProgressControl_init_value:
        
        """
        The current value.
        """

        return self._value
    
    @property
    def init_time(self):

        """
        The time spot of the first value.
        """

        return self._init_time
    
    @property
    def last_time(self):

        """
        The time spot of the final value.
        """

        return self._last_time

    def move(self, 
             size: _type_MultiLineProgressControl_move_size):
        
        """
        Move the internal value.
        
        :param size:
            The amount to move it by.
        """

        last_time = self._last_time = time.perf_counter()

        if self._init_time is None:
            self._init_time = last_time
        
        self._value += size

    def get_line(self, 
                 width: _type_MultiLineProgressControl_get_line_width) -> str:
        
        """
        Get the current line text.
        
        :param width:
            The maximum amount of allowed characters.
        """

        ratio = width / self._total
        value = min(width, self._value * ratio)

        full_count = math.floor(value)

        result = self._runes[-1] * full_count

        if full_count < width:
            suffix_index = (value - full_count) * len(self._runes)
            suffix_index = math.floor(suffix_index)
            suffix = self._runes[suffix_index]
            result = result + suffix

        result = _helpers.paint_text(self._color, result)

        return result
    
    def _get_info_made(self):

        if self._init_time is None:
            return None

        suffix = _helpers.get_or_call(self._suffix, self)

        if suffix is None:
            return None

        total_deno, total_unit = self._denominate(self._total)

        total_text = f'{self._total / total_deno:0.0f}'
        value_text = f'{self._value / total_deno:0.0f}'
        
        elapse = time.perf_counter() - self._init_time
        elapse_text = _helpers.format_seconds(elapse, depth = 2)

        speed = self._value / elapse
        speed_deno, speed_unit = self._denominate(speed)
        speed_text = f'{speed / speed_deno:0.2f}'

        remain = (self._total - self._value) / speed
        remain_text = _helpers.format_seconds(remain, depth = 2)

        return suffix.format(
            value = value_text, total = total_text, total_unit = total_unit,
            speed = speed_text, speed_unit = speed_unit, 
            elapse = elapse_text, remain = remain_text
        )
    
    def _get_info_done(self):
        
        epilogue = _helpers.get_or_call(self._epilogue, self)
        
        if epilogue is None:
            return self._get_info_made()
        
        return epilogue
    
    def _get_info(self):
        
        if self._value < self._total:
            info = self._get_info_made()
        else:
            info = self._get_info_done()

        if not info is None:
            info = _helpers.paint_text(self._color, info)
        
        return info
    
    def get_info(self) -> str:
        
        """
        Get the current info text.
        
        Will be an empty string if the internal deque is empty
        """
        
        return self._get_info()
        

@_theme.add('graphics.LineProgress.prefix')
def _get_MultiLineProgress_prefix(self):
    
    pass


@_theme.add('graphics.LineProgress.suffix')
def _get_MultiLineProgress_suffix(progress, delimit = ' '):
    
    infos = (control.get_info() for control in progress.controls)
    infos = (info for info in infos if not info is None)
    
    return ' ' + delimit.join(infos)


_type_MultiLineProgress_init_controls    = typing.List[MultiLineProgressControl]
_type_MultiLineProgress_init_width       = _type_MultiLineProgressControl_get_line_width
_type_MultiLineProgress_init_empty       = str
_type_MultiLineProgress_init_prefix      = _type_init_Graphic_prefix
_type_MultiLineProgress_init_suffix      = _type_init_Graphic_suffix
_type_MultiLineProgress_init_prefix_wall = str
_type_MultiLineProgress_init_suffix_wall = str


class MultiLineProgress(Graphic):
    
    """
    A manager for progress controls.
    
    :param controls:
        The controls to manage.
    :param width:
        The default width for each line.
    :param empty:
        The rune to fill empty space with.
    :param prefix_wall:
        Added before the line.
    :param prefix_wall:
        Added after the line.

    .. code-block:: python

        total = 1600
        controls = [
            survey.graphics.MultiLineProgressControl(total, color = survey.colors.basic('blue' ), epilogue = 'done!'),
            survey.graphics.MultiLineProgressControl(total, color = survey.colors.basic('red'  )),
            survey.graphics.MultiLineProgressControl(total, color = survey.colors.basic('green'), epilogue = lambda context: 'dynamic done!')
        ]
        # lower-index controls take longer to iterate, but speed up as the higher-index controls complete 
        with survey.graphics.MultiLineProgress(controls, prefix = 'Loading '):
            for _ in range(total):
                for index, control in enumerate(controls, start = 1):
                    if control.value >= total:
                        continue
                    for _ in range(index):
                        control.move(1)
                        time.sleep(0.01)
    
    |theme| :code:`graphics.MultiLineProgress`

    .. image:: /_static/images/graphics.multilineprogress-1.gif
    """
    
    __slots__ = ('_controls',)

    _interface_frequency = 0.1

    @_theme.add('graphics.MultiLineProgress')
    def __init__(self, 
                 controls   : _type_MultiLineProgress_init_controls, 
                 *args, 
                 width      : _type_MultiLineProgress_init_width       = 50,
                 empty      : _type_MultiLineProgress_init_empty       = ' ',
                 prefix     : _type_MultiLineProgress_init_prefix      = _get_MultiLineProgress_prefix,
                 suffix     : _type_MultiLineProgress_init_suffix      = _get_MultiLineProgress_suffix,
                 prefix_wall: _type_MultiLineProgress_init_prefix_wall = '|',
                 suffix_wall: _type_MultiLineProgress_init_suffix_wall = '|',
                 **kwargs):
        
        self._controls = controls

        def get(*args):
            main_line = list(empty * width)
            controls = (control for control in self._controls if not control.init_time is None)
            rank = lambda control: (control.value / control.total, - control.last_time)
            for control in sorted(controls, key = rank, reverse = True):
                rest_text = control.get_line(width)
                rest_line = _helpers.split_line(rest_text)
                main_line[0:len(rest_line)] = rest_line
            return prefix_wall + ''.join(main_line) + suffix_wall
        
        super().__init__(get, *args, prefix = prefix, suffix = suffix, **kwargs)
        
    @property
    def controls(self) -> _type_MultiLineProgress_init_controls:
        
        """
        The internal controls.
        
        Can be modified in any way at any time.
        """
        
        return self._controls
    
    
_type_LineProgress_init_total = _type_MultiLineProgressControl_init_total

_type_LineProgress_from_iterable_iterable = collections.abc.Container
_type_LineProgress_from_iterable_total    = typing.Union[_type_MultiLineProgressControl_init_total, None]
_type_LineProgress_from_iterable_count    = typing.Union[typing.Callable[[typing.Any], int], None]

_type_LineProgress_move_size = _type_MultiLineProgressControl_move_size


class LineProgress(MultiLineProgress):
    
    """
    A simple line progress with one internal control.
    
    :param total:
        The total expected value.

    .. code-block:: python

        total = 1600
        numbers = list(range(total))
        for number in survey.graphics.LineProgress.from_iterable(numbers, prefix = 'Loading ', epilogue = 'done!'):
            time.sleep(0.01)

    .. code-block:: python

        total = 1600
        numbers = list(range(total))
        with survey.graphics.LineProgress(len(numbers), prefix = 'Loading ', epilogue = 'done!') as progress:
            for number in numbers:
                time.sleep(0.01)
                progress.move(1)
    
    |theme| :code:`graphics.LineProgress`

    .. image:: /_static/images/graphics.lineprogress-1.gif
    """
    
    __slots__ = ()
    
    @_theme.add('graphics.LineProgress')
    def __init__(self, 
                 total: _type_LineProgress_init_total, 
                 *args, 
                 **kwargs):
        
        control_kwargs_names = _helpers.get_function_args_names(MultiLineProgressControl)
        control_kwargs = _helpers.yank_dict(kwargs, control_kwargs_names)
        control = MultiLineProgressControl(total, **control_kwargs)
        controls = (control,)
        
        super().__init__(controls, *args, **kwargs)
        
    @classmethod
    def from_iterable(cls, 
                      iterable: _type_LineProgress_from_iterable_iterable, 
                      *args, 
                      total   : _type_LineProgress_from_iterable_total = None, 
                      count   : _type_LineProgress_from_iterable_count = None, 
                      **kwargs):
        
        """
        Yields values from the iterable and displays the progress to completion.
        
        :param iterable:
            The iterable to yield values from.
        :param total:
            The total of the iterable if it cannot be determined directly from it.
        :param count:
            Can be used to specify by how much to increase the internal value per yield (default: ``1``).
            
        :raise TypeError: The iterable does not implement ``__len__`` and :paramref:`.total` has not been used.
        """

        if total is None:
            try:
                total = len(iterable)
            except TypeError as error:
                try:
                    total = iterable.__length_hint__()
                except AttributeError:
                    raise error
            
        if count is None:
            count = lambda value: 1
        
        with cls(total, *args, **kwargs) as self:
            for data in iterable:
                size = count(data)
                self._move(size)
                yield data

    @classmethod
    def from_response(cls, response, *args, **kwargs):

        """
        Yields chunks of data as received from the response.
        """

        total = int(response.headers['Content-length'])

        iterable = response.iter_content()
        
        units = ('B', 'KB', 'MB', 'GB')
        basic = 10 ** 3

        def denominate(value):
            for power, unit in reversed(tuple(enumerate(units))):
                ratio = basic ** power
                check = value / ratio / 10
                if check > 1:
                    break
            return (ratio, unit)
        
        def count(data):
            return len(data)

        return cls.from_iterable(iterable, *args, total = total, denominate = denominate, count = count, **kwargs) 
        
    def _move(self, size):
        
        self._controls[0].move(size)
        
    def move(self, 
             size: _type_LineProgress_move_size):
        
        """
        Move the underlying control's value.
        
        :param size:
            Same as :paramref:`.MultiLineProgressControl.move`.
        """
        
        self._move(size)
        
        



"""
Utilities for a :class:`~.visuals.Visual` that facilitates the main execution ground 
for  class:`.~widgets.Widget` by providing dynamic chunks of information around the screen.
"""

import typing
import copy

from . import _helpers
from . import _colors
from . import _visuals
from . import _funnels
from . import _system


__all__ = ('get', 'warn')


_warn_lines = [[]]


def _funnel_text_linesep(force, pre):

    def funnel(lines, point):
        if not force and not len(lines) > 1 and not any(lines):
            return
        if pre:
            lines.insert(0, [''])
            point[0] += 1
        else:
            lines.append([''])
    
    return funnel


_type_get_multi_maybe = bool
_type_get_multi_force = bool
_type_get_site        = str
_type_get_info_get    = _visuals._type_Text_init_get
_type_get_hint_get    = _visuals._type_Text_init_get
_type_get_body_get    = _visuals._type_Text_init_get
_type_get_info_color  = str
_type_get_hint_color  = str
_type_get_warn_color  = str


def _get(multi_maybe: _type_get_multi_maybe,
         multi_force: _type_get_multi_force, 
         site       : _type_get_site,
         info_get   : _type_get_info_get,
         hint_get   : _type_get_hint_get,
         body_get   : _type_get_body_get,
         info_color : _type_get_info_color = _colors.basic('cyan'), 
         hint_color : _type_get_hint_color = _colors.basic('black'), 
         warn_color : _type_get_warn_color = _colors.basic('red')):

    """
    Create a stage visual.

    :param multi_maybe:
        Whether the body may start on a separate line from the info and hint.
    :param multi_force:
        Whether the body has to start on a separate line from the info and hint.
    :param site:
        Can be :code:`'info'` or :code:`'body'` to place the cursor on the info or body respectively.
    :param info_get:
        Used to fetch info ``(lines, point)``.
    :param hint_get:
        Used to fetch hint ``(lines, point)``.
    :param body_get:
        Used to fetch body ``(lines, point)``.
    :param info_color:
        The color to paint info with.
    :param hint_color:
        The color to paint hint with.
    :param warn_color:
        The color to paint warn with.
    """

    global _warn_lines
    
    warn_lines = _warn_lines = [[]]
    
    info_funnel_leave_group = []
    info_funnel_leave_entry = _funnels.text_paint(info_color)
    info_funnel_leave_group.append(info_funnel_leave_entry)
    def info_funnel_leave_entry(lines, point):
        if not site == 'info' and not any(lines):
            return
        _funnels.text_bloat_horizontal.call(_funnels.JustType.start, 1, ' ', lines, point)
    info_funnel_leave_group.append(info_funnel_leave_entry)
    info_funnel_leave = _helpers.chain_functions(*info_funnel_leave_group)
    info_visual = _visuals.Text(info_get, funnel_leave = info_funnel_leave)

    hint_funnel_leave_group = []
    hint_funnel_leave_entry = _funnels.text_paint(hint_color)
    hint_funnel_leave_group.append(hint_funnel_leave_entry)
    hint_funnel_leave = _helpers.chain_functions(*hint_funnel_leave_group)
    hint_visual = _visuals.Text(hint_get, funnel_leave = hint_funnel_leave)

    warn_funnel_leave_group = []
    warn_funnel_leave_entry = _funnels.text_paint(warn_color)
    warn_funnel_leave_group.append(warn_funnel_leave_entry)
    warn_funnel_leave = _helpers.chain_functions(*warn_funnel_leave_group)
    warn_visual = _visuals.Text.link(warn_lines, [0, 0], funnel_leave = warn_funnel_leave)

    info_visual_get = info_visual.get
    hint_visual_get = hint_visual.get

    head_visual_tiles = {(0, 0): info_visual_get, (0, 1): hint_visual_get}
    head_funnel_enter_group = []
    head_funnel_entry_entry = _funnels.mesh_grid_fill()
    head_funnel_enter_group.append(head_funnel_entry_entry)
    head_funnel_enter = _helpers.chain_functions(*head_funnel_enter_group)
    head_funnel_leave_group = []
    if multi_maybe:
        head_funnel_leave_entry = _funnel_text_linesep(multi_force, False)
        head_funnel_leave_group.append(head_funnel_leave_entry)
    head_funnel_leave = _helpers.chain_functions(*head_funnel_leave_group)
    head_visual = _visuals.Mesh.link(head_visual_tiles, [0, 0], funnel_enter = head_funnel_enter, funnel_leave = head_funnel_leave)

    body_visual = _visuals.Text(body_get)

    foot_visual_tiles = {(0, 0): warn_visual.get}
    foot_funnel_enter = head_funnel_enter
    foot_funnel_leave_group = []
    foot_funnel_leave_entry = _funnel_text_linesep(False, True)
    foot_funnel_leave_group.append(foot_funnel_leave_entry)
    foot_funnel_leave = _helpers.chain_functions(*foot_funnel_leave_group)
    foot_visual = _visuals.Mesh.link(foot_visual_tiles, [0, 0], funnel_enter = foot_funnel_enter, funnel_leave = foot_funnel_leave)

    head_visual_get = head_visual.get
    body_visual_get = body_visual.get
    foot_visual_get = foot_visual.get

    main_visual_tiles = [head_visual_get, body_visual_get, foot_visual_get]

    main_visual_point_latch_group = {
        'info': head_visual_get,
        'body': body_visual_get
    }

    main_visual_point_latch = main_visual_point_latch_group[site]
    main_visual_point_index = main_visual_tiles.index(main_visual_point_latch)
    main_visual_point = [main_visual_point_index]

    main_funnel_leave_group = []
    # main_funnel_leave_entry = _funnels.text_max_dynamic(_system.cursor.measure)
    # main_funnel_leave_group.append(main_funnel_leave_entry)
    main_funnel_leave = _helpers.chain_functions(*main_funnel_leave_group)

    main_visual = _visuals.Line.link(main_visual_tiles, main_visual_point, funnel_leave = main_funnel_leave)

    return main_visual


get = _get


_type_warn_lines = _visuals._type_Text_link_lines


def _warn(lines: _type_warn_lines):

    """
    Create a warning udnerneath the body.

    :param lines:
        The warning lines.
    """

    _warn_lines[:] = copy.deepcopy(lines)


warn = _warn


import itertools

from . import _helpers
from . import _ansi
from . import _io


__all__ = ('Intel',)


class _EmptyRead(Exception):

    __slots__ = ()


_check_any = lambda *args, **kwargs: True


class Intel:

    def __init__(self, 
                 io: _io.IO):

        self._io = io

        self._rune_buffer = []
        self._code_buffer = []

    @property
    def io(self):

        return self._io

    def _fill_text(self):

        if self._rune_buffer:
            return
        
        text = self._io.recv()

        if text == '\x1b':
            return

        self._rune_buffer.extend(text)

    def _read_text(self):

        self._fill_text()

        try:
            rune = self._rune_buffer.pop(0)
        except IndexError:
            raise _EmptyRead()

        return rune
    
    _empty = _ansi.Escape('')
    
    def _fill_code(self):
        
        try:
            code = _ansi.parse(self._read_text)
        except _EmptyRead:
            code = self._empty

        self._code_buffer.append(code)
    
    @_helpers.ctxmethod(lambda self: self._io)
    def _read_code(self, check):

        if check is None:
            check = _check_any

        for index in itertools.count(0):
            for _ in range(2):
                try:
                    code = self._code_buffer[index]
                except IndexError:
                    self._fill_code()
                else:
                    break
            if check(code):
                break
            index += 1

        del self._code_buffer[index]
        
        return code
    
    def _read(self, check):

        code = self._read_code(check)

        return code
    
    def read(self, check = None):

        return self._read(check)
    


import collections

from . import _core


__all__ = ()


Control = collections.namedtuple('Control', 'event function')


def _control(event):

    def decorator(function):
        return Control(event, function)

    return decorator


get = _control


@_control(_core.Event.insert)
def text_insert(mutate, info):

    runes = (info.rune,)

    mutate.insert(runes)

    size = len(runes)

    mutate.move_x(size)


@_control(_core.Event.arrow_left)
def text_move_left(mutate, info):

    size = - 1

    mutate.move_x(size)


@_control(_core.Event.arrow_right)
def text_move_right(mutate, info):

    size = 1

    mutate.move_x(size)


@_control(_core.Event.arrow_up)
def text_move_up(mutate, info):

    size = - 1

    mutate.move_y(size)


@_control(_core.Event.arrow_down)
def text_move_down(mutate, info):

    size = 1

    mutate.move_y(size)


@_control(_core.Event.delete_left)
def text_delete_left(mutate, info):

    size = 1

    mutate.move_x(- size)
    mutate.delete(size)


@_control(_core.Event.delete_right)
def text_delete_right(mutate, info):

    size = 1

    mutate.delete(size)


def done(lines, point):

    cur_y = point[0]
    fin_y = len(lines) - 1
    
    return cur_y == fin_y and len(lines) > 2 and not any(lines[-2:])


@_control(_core.Event.enter)
def text_newline(mutate, info):
    lines = mutate.lines
    point = mutate.point
    if done(lines, point):
        lines[-2:] = ()
        y = len(lines) - 1
        x = len(lines[y])
        mutate.point[:] = (y, x)
        raise _core.Terminate()
    mutate.newline()


@_control(_core.Event.arrow_left)
def mesh_move_left(mutate, info):

    direction = 180

    mutate.move(direction)


@_control(_core.Event.arrow_right)
def mesh_move_right(mutate, info):

    direction = 0

    mutate.move(direction)


@_control(_core.Event.arrow_up)
def mesh_move_up(mutate, info):

    direction = + 90

    mutate.move(direction)


@_control(_core.Event.arrow_down)
def mesh_move_down(mutate, info):

    direction = - 90

    mutate.move(direction)


@_control(_core.Event.arrow_up)
def mesh_move_up_reverse(mutate, info):

    direction = - 90

    mutate.move(direction)


@_control(_core.Event.arrow_down)
def mesh_move_down_reverse(mutate, info):

    direction = + 90

    mutate.move(direction)


@_control(_core.Event.insert)
def mesh_insert(mutate, info):

    runes = (info.rune,)

    mutate.search_insert(runes)


@_control(_core.Event.delete_left)
def mesh_delete(mutate, info):

    size = 1

    mutate.search_delete(size)


@_control(_core.Event.enter)
def mesh_enter(mutate, info):

    raise _core.Terminate()




import typing
import enum
import threading

from . import _helpers
from . import _ansi
from . import _intel


__all__ = ('Source', 'Event')


class EventType(enum.Enum):

    none     = None
    text     = _ansi.Text
    control  = _ansi.Control
    escape   = _ansi.Escape
    sequence = _ansi.Sequence


class Event(str, enum.Enum):

    """
    Event()

    Names for received events.
    """

    #: 
    insert       = 'insert'
    #: 
    arrow_left   = 'arrow_left'
    #: 
    arrow_right  = 'arrow_right'
    #: 
    arrow_up     = 'arrow_up'
    #: 
    arrow_down   = 'arrow_down'
    #: 
    delete_left  = 'delete_left'
    #: 
    delete_right = 'delete_right'
    #: 
    escape       = 'escape'
    #: 
    indent       = 'indent'
    #: 
    enter        = 'enter'


class _EmptyRead(Exception):

    __slots__ = ()


_type_Source_init_callback = typing.Callable[[Event, _ansi._type_parse_return], None]


class Source:

    """
    Translates incoming ANSI sequences to events and calls back with relevant information.

    :param callback:
        Called with ``(name, info)`` upon receiving and translating.
    :param intel:
        Used for receiving from input.
    """

    __slots__ = ('_callback', '_intel', '_lock')

    def __init__(self, 
                 callback: _type_Source_init_callback, 
                 intel   : _intel.Intel):

        self._callback = callback
        self._intel = intel

        self._lock = threading.RLock()

    @property
    def intel(self):

        return self._intel

    _group = {
        EventType.control: {
            '\x7f': Event.delete_left,
            '\x08': Event.delete_right,
            '\x0d': Event.enter,
            '\x09': Event.indent,
            '\x10': Event.arrow_up,
            '\x0E': Event.arrow_down
        },
        EventType.escape: {
            '': Event.escape
        },
        EventType.sequence: {
            'A': Event.arrow_up,
            'B': Event.arrow_down,
            'D': Event.arrow_left,
            'C': Event.arrow_right
        }
    }

    _mimic = {
        EventType.control: {
            '\n': _ansi.parse_iterable('\x0d')
        },
        EventType.escape: {
            'f': _ansi.parse_iterable('\x1b[1;3C'),
            'b': _ansi.parse_iterable('\x1b[1;3D')
        },
        EventType.sequence: {
            '~': _ansi.parse_iterable('\x08')
        }
    }

    def _process(self, info):

        while True:
            type = EventType(info.__class__)
            try:
                info = self._mimic[type][info.rune]
            except KeyError:
                break

        try:
            names = self._group[type]
        except KeyError:
            name = Event.insert
        else:
            try:
                name = names[info.rune]
            except KeyError:
                return
        
        self._callback(name, info)

    @_helpers.ctxmethod(lambda self: self._lock)
    def _advance(self):

        info = self._intel.read()

        self._process(info)

    def _listen(self):

        while True:
            self._advance()

    def listen(self):

        """
        Begin listening for events.
        """

        self._listen()


import types
import functools
import operator
import re
import math
import itertools
import collections
import inspect
import copy
import contextlib
import statistics

from . import _constants


from ._core._helpers import *


class InfoErrorMixin:

    __slots__ = ()

    def __init__(self, text, **info_kwargs):

        info = types.SimpleNamespace(**info_kwargs)

        super().__init__(text, info)

    @property
    def text(self):

        return self.args[0]

    @property
    def info(self):

        return self.args[1]


def noop(*args, **kwargs):

    pass


@contextlib.contextmanager
def noop_contextmanager(*args, **kawrgs):

    yield


def text_point_to_index(lines, cy, cx):

    lines = lines[:cy]

    ci = sum(map(len, lines)) + len(lines) + cx

    return ci


def text_index_to_point(lines, ci):

    cy = 0
    cx = ci
    for sx in map(len, lines):
        sx += 1
        ci -= sx
        if ci < 0:
            break
        cy += 1
        cx = ci

    return (cy, cx)


def get_point_neighbors(origin, radius):

    steps = range(- radius, radius + 1)

    for i, j in itertools.product(steps, steps):
        if not radius in map(abs, (i, j)):
            continue
        yield (origin[0] + i, origin[1] + j)


def get_point_distance_to_point(origin, target):

    return math.sqrt(sum((px - qx) ** 2.0 for px, qx in zip(origin, target)))

def get_point_distance_to_line(slope, intercept, target):

    return abs(-slope * target[1] + target[0] - intercept) / math.hypot(slope, 1)


def get_point_direction_to_point(origin, target):

    dy = target[0] - origin[0]
    dx = target[1] - origin[1]
    
    angle = math.atan2(dy, dx)
    direction = math.degrees(angle)

    return direction


def get_point_directional(decide, origin, direction, targets, window = 90, ignore_point = False):

    angle = math.radians(direction % 180)
    slope = math.tan(angle)
    intercept = origin[0] - slope * origin[1]

    def check_direction(target):
        if target == origin:
            return False
        cur_direction = direction
        new_direction = get_point_direction_to_point(origin, target)
        return abs(new_direction - cur_direction) < window

    targets = tuple(filter(check_direction, targets))

    def get_score(target):
        distance_to_line = get_point_distance_to_line(slope, intercept, target)
        distance_to_point = get_point_distance_to_point(origin, target)
        if ignore_point:
            distance_to_line *= - 1
        return distance_to_line + distance_to_point

    target = decide(targets, key = get_score, default = origin) 

    return target
    

def get_point_directional_next(origin, direction, radius):

    points = tuple(get_point_neighbors(origin, radius))

    return get_point_directional(min, origin, direction, points)


def reverse_direction(direction):

    direction = direction % 360 - 180

    return direction if direction > - 180 else direction + 360


class SGR:

    _cre = re.compile('((?:\x1b[\\[0-?]*m)+)')

    @classmethod
    def split(cls, value):

        return cls._cre.split(value)

    @classmethod
    def yank(cls, value):

        chunks = cls.split(value)

        for index, chunk in enumerate(chunks):
            if index % 2:
                continue
            yield chunk

    @classmethod
    def clean(cls, value):

        chunks = cls.yank(value)

        return functools.reduce(operator.add, chunks)
    
    @classmethod
    def apply(cls, function, value):

        chunks = cls.split(value)

        for index, chunk in enumerate(chunks):
            if index % 2:
                continue
            chunks[index] = function(chunk)

        return ''.join(chunks)
    
    resets = {
        '\x1b[0m', 
        '\x1b[22m ','\x1b[23m', '\x1b[24m', '\x1b[25m', '\x1b[26m',' \x1b[27m','\x1b[28m',
        '\x1b[39m', '\x1b[49m', '\x1b[50m', '\x1b[54m', '\x1b[55m', '\x1b[59m', '\x1b[65m', '\x1b[75m'
    }


def chain_functions(*functions):

    functions = tuple(filter(callable, functions))

    def function(*args, **kwargs):
        for function in functions:
            function(*args, **kwargs)

    return function


def split_line(value):

    chunks = SGR.split(value)

    buffer = list(chunks.pop(0))

    for index, chunk in enumerate(chunks):
        if not index % 2:
            continue
        if chunk:
            runes = list(chunk)
            attach = chunks[index - 1]
            try:
                if attach in SGR.resets:
                    try:
                        buffer[- 1] += attach
                    except IndexError:
                        pass
                    else:
                        continue
                runes[0] = attach + runes[0]
            finally:
                buffer.extend(runes)
        else:
            # last chunk is always empty
            try:
                buffer[- 1] += chunks[index - 1]
            except IndexError:
                pass

    return buffer


def split_lines(value, *args, **kwargs):

    values = value.split(_constants.linesep)

    get = lambda value: split_line(value, *args, **kwargs)

    lines = list(map(get, values))

    return lines


def join_line(line):

    value = ''.join(line)

    return value


def join_lines(lines, *args, **kwargs):

    get = functools.partial(join_line, *args, **kwargs)

    values = map(get, lines)

    value = _constants.linesep.join(values)

    return value


class IdentifiableSGR(str):

    __slots__ = ('_names',)

    def __new__(cls, value, **names):

        self = super().__new__(cls, value)

        self._names = names

        return self
    
    def describe(self):

        parts = ((name, value) for name, value in self._names.items() if value)

        description = ' '.join(map('{0[0]}:{0[1]}'.format, parts))

        return description
    
    def __repr__(self):

        description = self.describe()

        return f'<{description}>'


def paint_rune(color, rune, reset = '\x1b[0m'):

    if color is None:
        return rune

    return color + rune + reset


class Graffity(str):

    __slots__ = ('_color', '_value')

    def __new__(cls, color, value):

        if not isinstance(color, IdentifiableSGR):
            return value

        self = super().__new__(cls, value)

        self._color = color
        self._value = value

        return self

    def __repr__(self):

        color_description = self._color.describe()
        value_description = repr(self._value)

        return f'<{color_description} {value_description}>'


def paint_text(color, value, *args, **kwargs):

    paint = functools.partial(paint_rune, color)

    value = ''.join(map(paint, value))

    return Graffity(color, value)


def paint_line(color, line, *args, **kwargs):

    for index, rune in enumerate(line):
        line[index] = paint_rune(color, rune, *args, **kwargs)


def paint_lines(color, lines, *args, **kwargs):

    for line in lines:
        paint_line(color, line, *args, **kwargs)


def merge_lines(main, *rest):

    main = copy.deepcopy(main)
    for rest in map(copy.deepcopy, rest):
        main[- 1].extend(rest.pop(0))
        main.extend(rest)

    return main


def check_lines(lines):

    return len(lines) > 1 or any(lines)


def squeeze_spots(axis, spots, dimensions = None, start = 0):

    if dimensions is None:
        dimensions = len(next(iter(spots)))

    oths = (axis - 1) % dimensions

    buckets = collections.defaultdict(list)

    for spot in spots:
        cur_a = spot[axis]
        cur_o = spot[oths]
        buckets[cur_a].append(cur_o)

    all_a = range(0, len(buckets))

    for new_a, cur_info in zip(all_a, buckets.items()):
        cur_a, all_o = cur_info
        for new_o, cur_o in enumerate(all_o, start = start):
            cur_spot = [cur_o] * 2
            cur_spot[axis] = cur_a
            new_spot = [new_o] * 2
            new_spot[axis] = new_a
            yield tuple(map(tuple, (cur_spot, new_spot)))


def decorify(function):

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        return functools.partial(function, *args, **kwargs)
    
    wrapper.call = function

    wrapper.__annotations__ = function.__annotations__

    return wrapper


def get_axis_point(dimensions, axis, default, index):

    point = [default] * dimensions
    
    point[axis] = index

    return point


@functools.lru_cache()
def get_function_parameters(function):

    if isinstance(function, type):
        functions = function.__mro__
    else:
        functions = ()

    signatures = reversed(tuple(map(inspect.signature, functions)))

    ignore_kinds = {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}

    parameters = {}
    for signature in signatures:
        parameters.update(signature.parameters)

    parameters = {name: parameter for name, parameter in parameters.items() if not parameter.kind in ignore_kinds}

    return parameters


@functools.lru_cache()
def get_function_args_names(function):

    parameters = get_function_parameters(function)

    return tuple(parameters)


def get_function_args_default(function, name):

    parameters = get_function_parameters(function)

    parameter = parameters[name]

    default = parameter.default

    if default is inspect.Parameter.empty:
        raise ValueError(f'parameter "{name}" of {function} has no default')

    return default


def get_function_arg_safe(function, name, store, pop = False):

    getter = store.pop if pop else store.__getitem__

    try:
        return getter(name)
    except KeyError:
        pass

    return get_function_args_default(function, name)


def yank_dict(store, keys):

    result = {}
    for key in keys:
        try:
            value = store.pop(key)
        except KeyError:
            continue
        result[key] = value

    return result


def split_list(store, check):

    result = ([], [])

    for value in store:
        result[check(value)].append(value)

    return result


def asset_value(value):

    lines = split_lines(value)

    point_y = len(lines) - 1
    point_x = len(lines[point_y])
    point = [point_y, point_x]

    return (lines, point)


def get_or_call(value, *args, **kwargs):
    
    return value(*args, **kwargs) if callable(value) else value


auto = type('auto', (), {'__slots__': (), '__repr__': lambda self: self.__class__.__name__})()


def format_seconds(value, delimit = ':', fill = '0', depth = None):

    depth = max(0, depth - 1)

    dividers = (60, 60, 24)
    segments = []

    value = round(value)

    for divider_index, divider in enumerate(dividers):
        top, sub = divmod(value, divider)
        sub = str(sub)
        sub = fill * (len(str(divider)) - len(sub)) + sub
        segments.append(sub)
        if not top and (depth is None or not divider_index < depth):
            break
        value = top

    return delimit.join(reversed(segments))


import typing

from . import _screen
from . import _render
from . import _handle
from . import _io
from . import _cursor
from . import _ansi


__all__ = ('SkipDraw', 'Console')


class SkipDraw(Exception):

    """
    Signals the invokation loop to not skip drawing for the current turn.
    """

    __slots__ = ()


class Console:

    """
    Standardizes interaction with the IO into a wait-receive-invoke loop.

    :param handle:
        Used for the invokation loop.
    :param screen:
        Used for visualizing.
    """

    __slots__ = ('_handle', '_screen')

    def __init__(self, 
                 handle: _handle.Handle, 
                 screen: _screen.Screen):

        self._handle = handle
        self._screen = screen

    @property
    def handle(self) -> _handle.Handle:

        """
        The internal handle.
        """

        return self._handle
    
    @property
    def screen(self) -> _screen.Screen:

        return self._screen
    
    def _start(self, sketch, invoke):
        
        def print(re):
            self._screen.print(sketch, re)

        def callback(name, info):
            try:
                invoke(name, info)
            except SkipDraw:
                return
            print(True)

        print(False)

        self._handle.start(callback)

    def start(self,
              sketch: _screen._type_Screen_draw_sketch, 
              invoke: _handle._type_Handle_start_invoke):
        
        """
        Creates an invokation-visualization loop. 
        
        1. ``sketch`` is used for the initial print.
        2. ``invoke`` is used once input is received. 
        3. ``sketch`` is used for the updated print.
        4. Repeat from step **2** until :exc:`.handle.Terminate` gets raised.

        :param sketch:
            Used for visualizing.
        :param invoke:
            Called upon receiving info from the IO.
        """
        
        self._start(sketch, invoke)


import typing
import math
import functools
import dataclasses
import threading
import weakref
import unicodedata

from . import _helpers
from . import _constants
from . import _ansi
from . import _cursor
from . import _intel


__all__ = ('Render',)


def _get_line_wrap_y(limit: int, 
                     size: int):

    """
    Get the amount of y-space needed to render a line in a given x-space.

    :param limit:
        The x-space limit.
    :param size:
        The size of the line.

    .. code-block:: python

        _get_line_wrap_y(5, 8 ) # 2
        _get_line_wrap_y(9, 4 ) # 1
        _get_line_wrap_y(5, 33) # 7
    """

    return math.floor(size / limit) + 1


def _get_line_wrap_y_sum(limit: int, 
                         sizes: typing.List[int]):

    """
    Get the amount of y-space needed to render a list of lines in a given x-space.

    :param limit:
        The x-space limit.
    :param size:
        The size of the line.

    .. code-block:: python

        _get_line_wrap_y(5, (8, 4)) # 3
        _get_line_wrap_y(9, (4, 2)) # 2
        _get_line_wrap_y(5, (6, 9)) # 4
    """

    get = functools.partial(_get_line_wrap_y, limit)

    return sum(map(get, sizes))


def _get_point_wrap(limit: int, 
                    sizes: typing.List[int], 
                    cur_y: int, 
                    cur_x: int):

    """
    Get the (y, x) in a list of lines after accounting for a given x-space.
    """

    req_x = _get_line_wrap_y(limit, cur_x)

    usr_y = _get_line_wrap_y_sum(limit, sizes[:cur_y]) + req_x
    usr_x = cur_x - (req_x - 1) * limit

    # turn to index via count
    usr_y -= 1

    return (usr_y, usr_x)


class Memory:

    __slots__ = ('y', 'x', '__weakref__')

    def __init__(self, y, x):

        self.y = y
        self.x = x
    

_type_Render_draw_lines = typing.List[typing.List[str]]
_type_Render_draw_point = typing.Union[typing.List[int], None]


class Render:

    """
    Assists with printing text in a predictible way.

    .. note::
        In contrast to how :class:`.Cursor` works, point coordinates here are 0-offset indexes.
    
    :param cursor:
        Used for moving and clearing.
    :param intel:
        Used for sending to the output.

    .. code-block:: python
    
        io = IO(sys.stdin, sys.stdout)
        intel = Intel(io)
        cursor = Cursor(intel)
        render = Render(cursor, intel)
        # draw "Hello\\nWorld" and place cursor between "r" and "l".
        render.draw([['H', 'e', 'l', 'l', 'o'], ['W', 'o', 'r', 'l', 'd']], [1, 3])
    """

    _history = weakref.WeakSet()

    _style_reset = _ansi.get_control('m', 0)

    __slots__ = ('_cursor', '_memory', '_lock')

    def __init__(self,
                 cursor: _cursor.Cursor):

        self._cursor = cursor
        self._memory = None
        
        self._lock = threading.Lock()
    
    @property
    def cursor(self) -> _cursor.Cursor:

        """
        The cursor used for positioning the cursor and fetching coordinate information.
        """

        return self._cursor

    @property
    def memory(self) -> _type_Render_draw_point:

        """
        The last initial cursor position.
        """

        return self._memory
    
    def _send_direct(self, text):

        self._cursor.intel.io.send(text)

    def _send(self, clean, lines):

        for index, line in enumerate(lines):
            if index:
                self._send_direct(_constants.linesep)
            text = ''.join(line)
            self._send_direct(text)
            if clean:
                self._cursor.erase()

    def _move(self, sizes, cur_y, cur_x, max_x, point):

        usr_y, usr_x = point

        usr_x += cur_x if usr_y == 0 else 1

        usr_y, usr_x = _get_point_wrap(max_x, sizes, usr_y, usr_x)

        usr_y += cur_y
        
        self._cursor.move(usr_y, usr_x)

    @_helpers.ctxmethod(lambda self: self._lock)
    def _draw(self, learn, clean, lines, point):

        lines = tuple(map(tuple, lines))

        cur_y, cur_x = self._cursor.locate()
        max_y, max_x = self._cursor.measure()

        if not point is None:
            for rune in lines[point[0]][:point[1]]:
                try:
                    iuni = unicodedata.east_asian_width(rune)
                except TypeError:
                    continue
                cur_x += iuni in ('W', 'A', 'F')

        memory = Memory(cur_y, cur_x)

        if learn:
            self._memory = memory
            self._history.add(memory)

        self._send_direct(self._style_reset)

        self._send(True, lines)

        if clean:
            self._cursor.clear()

        sizes = list(map(len, lines))

        # adjust initial x
        sizes[0] += cur_x - 1

        # total y needed
        tot_y = _get_line_wrap_y_sum(max_x, sizes)
        # total y remaining 
        got_y = max_y - cur_y
        # total y generated
        ext_y = tot_y - got_y
        # might have enough
        ext_y = max(0, ext_y - 1)

        for oth_memory in self._history:
            # adjust extra y
            oth_memory.y -= ext_y

        if not point is None:
            # "memory" from loop above
            self._move(sizes, memory.y, memory.x, max_x, point)

    def draw(self, 
             lines: _type_Render_draw_lines,
             point: _type_Render_draw_point = None,
             *,
             clean: bool = True,
             learn: bool = True):

        """
        Draw lines and place the cursor among them.

        :param lines:
            The lines to draw.
        :param point:
            The coordiantes among the lines to place the cursor.
        :param clean:
            Whether to clear the rest of the screen after drawing the lines.
        :param learn:
            Whether to keep track of the initial cursor position.
        """

        self._draw(learn, clean, lines, point)

    def _back(self):

        memory = self._memory

        if memory is None:
            return
        
        self._cursor.move(memory.y, memory.x)

        self._memory = None

    def back(self):

        """
        Move the cursor to the position it was before drawing.
        """

        self._back()





import io

from . import _helpers


__all__ = ('IO',)


_system_module = _helpers.import_module_os('_io')


class BaseIO(_system_module.IO):

    __slots__ = ()

    def _ring(self):

        super()._send(b'\a')

    def ring(self):

        """
        Trigger the system's bell.
        """

        return self._ring()


class IO(BaseIO):

    """
    Manages the input and output buffers.

    .. code-block::

        io = IO(sys.stdin, sys.stdout)

        with io: 
            rune = io.recv()
            io.send('You pressed ' + rune)
    """

    __slots__ = ('_i_encoding', '_o_encoding')

    def __init__(self, 
                 i: io.TextIOWrapper, 
                 o: io.TextIOWrapper, 
                 *args, **kwargs):

        self._i_encoding = i.encoding
        self._o_encoding = o.encoding

        super().__init__(i.buffer, o.buffer, *args, **kwargs)

    def _encode(self, text):

        return text.encode(self._o_encoding)

    def _send(self, text):

        text_raw = self._encode(text)

        super()._send(text_raw)

    def _decode(self, text):

        return text.decode(self._i_encoding)

    def _recv(self):

        text_raw = super()._recv()

        text = self._decode(text_raw)

        return text

    def __iter__(self):

        iterable = super().__iter__()

        yield from map(self._decode, iterable) 



import termios

from . import _io_os


__all__ = ('IO',)


class IO(_io_os.StreamIO):

    __slots__ = ('_i_fd',)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._i_fd = self._i.fileno()

    def _wait(self, mode, switch):

        mode[6][termios.VMIN] = 1 if switch else 0

        return mode

    def _init(self):

        save = termios.tcgetattr(self._i_fd)

        mode = termios.tcgetattr(self._i_fd)

        mode[3] &= ~(termios.ECHO | termios.ECHONL | termios.ICANON)
        mode[6][termios.VTIME] = 0

        return (save, mode)

    def _set(self, mode):

        termios.tcsetattr(self._i_fd, termios.TCSAFLUSH, mode)




"""
Handling incoming events and their information by delegating to functions that can handle them.
"""

import typing
import enum

from . import _helpers
from . import _core
from . import _controls


__all__ = ('EventType', 'Abort', 'Handle')


class EventType(str, enum.Enum):

    """
    Flags whether the event is called back before or after invokation.
    """

    enter = 'enter'
    leave = 'leave'


class Abort(Exception):

    """
    Can be raised during a callback with :attr:`EventType.enter` to prevent invokation.
    """

    __slots__ = ()


_type_Handle_init_unsafe   = bool
_type_Handle_init_callback = typing.Callable[[typing.Tuple[EventType, _core.Event], _core._type_ansi_parse_return], None]

_type_Handle_add_control   = _controls.Control

_type_Handle_invoke_event  = _core.Event


class Handle:

    """
    Used for holding and calling "control" functions that determine action for each :class:`._core.Event`.

    :param unsafe:
        Whether :exc:`KeyError` should be surfaced when an invoke is missing.
    :param callback:
        Called before and after invokation with :attr:`.Event.enter.` and :attr:`.Event.leave` prepended in the arguments respectively.
    """

    __slots__ = ('_args', '_unsafe', '_controls', '_callback')

    def __init__(self, 
                 *args,
                 unsafe  : _type_Handle_init_unsafe   = False, 
                 callback: _type_Handle_init_callback = None):

        self._args = args
        self._unsafe = unsafe
        self._controls = {}
        self._callback = callback or _helpers.noop

    def add(self, 
            control: _type_Handle_add_control):

        """
        Add a control.
        """

        self._controls[control.event] = control

    def _dispatch(self, type, name, *args):
        
        event = (type, name)

        self._callback(event, *args)

    def invoke(self, 
               event: _type_Handle_invoke_event, 
               *args):

        """
        Invoke a control if it exists.
        """

        try:
            control = self._controls[event]
        except KeyError:
            if self._unsafe:
                raise
            return
        
        try:
            self._dispatch(EventType.enter, event, *args)
        except Abort:
            return

        control.function(*self._args, *args)

        self._dispatch(EventType.leave, event, *args)


        
        


import setuptools

with open('README.rst') as file:
    readme = file.read()

author = 'Exahilosys'
project = 'survey'

url = 'https://github.com/{0}/{1}'.format(author, project)

setuptools.setup(
    name = project,
    python_requires = '>=3.7',
    use_scm_version = True,
    setup_requires = [
        'setuptools_scm'
    ],
    url = url,
    packages = setuptools.find_packages(),
    license = 'MIT',
    description = 'A simple library for creating beautiful interactive prompts.',
    long_description = readme,
    extras_require = {
        'docs': [
            'sphinx',
            'sphinx-rtd-theme',
            'sphinx-paramlinks',
            'sphinx-autodoc-typehints<1.25.3'
        ]
    }
)


import msvcrt
import ctypes
import ctypes.wintypes

from . import _constants
from . import _io_os


__all__ = ('IO',)


_LINESEP = _constants.linesep.encode()


def _recv_linesep(text):

    return _LINESEP


def _recv_interrupt(text):

    raise KeyboardInterrupt


class IO(_io_os.StreamIO):

    __slots__ = ('_i_fh', '_o_fh', '_o_save')

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        i_fd = self._i.fileno()
        self._i_fh = msvcrt.get_osfhandle(i_fd)

        o_fd = self._o.fileno()
        self._o_fh = msvcrt.get_osfhandle(o_fd)

    def _wait(self, mode, switch):

        bit = 2

        if switch:
            # - ENABLE_LINE_INPUT
            mode &= ~ bit
        else:
            # + ENABLE_LINE_INPUT
            mode |= bit

        return mode

    @staticmethod
    def _getm(fh):

        c_mode = ctypes.wintypes.DWORD()

        ctypes.windll.kernel32.GetConsoleMode(fh, ctypes.byref(c_mode))

        mode = c_mode.value

        return mode

    def _init(self):

        save = self._getm(self._i_fh)

        # - ENABLE_PROCESSED_INPUT
        # - ENABLE_ECHO_INPUT
        # - ENABLE_INSERT_MODE
        # + ENABLE_VIRTUAL_TERMINAL_INPUT
        mode = save &~ (1 | 4 | 32) | 512

        return (save, mode)

    @staticmethod
    def _setm(fh, mode):

        c_mode = ctypes.wintypes.DWORD(mode)

        ctypes.windll.kernel32.SetConsoleMode(fh, c_mode)

    def _set(self, mode):

        self._setm(self._i_fh, mode)

    def _setup(self):

        mode = self._o_save = self._getm(self._o_fh)

        # + ENABLE_PROCESSED_OUTPUT
        # + ENABLE_VIRTUAL_TERMINAL_PROCESSING
        mode = mode | 1 | 4

        self._setm(self._o_fh, mode)

    def _reset(self):

        self._setm(self._o_fh, self._o_save)

    _recv_replace = {
        b'\r': _recv_linesep,
        b'\x03': _recv_interrupt
    }

    def _recv(self):

        text = super()._recv()

        while True:
            try:
                function = self._recv_replace[text]
            except KeyError:
                break
            else:
                text = function(text)
        
        return text

    def __enter__(self, *args):

        self._setup()

        return super().__enter__(*args)

    def __exit__(self, *args):

        self._reset()

        return super().__exit__(*args)